// ======================= MSP430F1xxx S/W   α׷ =================
//  CPU : MSP430F1121
//     ¥ : 2008-
//      : Ver1.0
//    : www.cpuplaza.com
// Ϸ : IAR-MSP430
// [ ]  : ACLK : 32.768Khz
//            MCLK : DCO(3.6864MHz)
//            SMCLK: DCO(3.6864MHz)
// ----------------------------------------------------------------------------
// ⺻ 
#include "msp430x11x1.h"
#include <stdio.h>

//  
#define		DELTA_1	900					// DCO = DELATA_1 * 4096 = 3.6864MHz
#define		DELTA_2	250					// DCO = DELATA_2 * 4096 = 1.024000MHz
#define		DELTA_3	64					// DCO = DELATA_3 * 4096 = 0.262144MHz

#define		DELTA	DELTA_1					// DCO = 3.6864MHz

//  
#define RXD	0x04						// RXD on P2.2
#define TXD	0x02						// TXD on P1.1

// Լ 
void cpu_setup(void);						// CPU ʱȭ
void Set_DCO(void);						// DCO 

void com_set(int baud);						//  
void rx_ready(void);						//  غ
void putchar_0(unsigned char datac);				//  1Ʈ ۽
void puts_0(unsigned char dbuf[]);				//  ڿ 

//  
unsigned int Bitime_5 = 192;					// 50us bit length 
unsigned int Bitime = 384;					// 104.1666us(9600Bps)

unsigned int TXData;						// ۽ 
unsigned int RXData;						//  
unsigned char TXBitCnt;						// ۽  ī
unsigned char RXBitCnt;						//   ī

// ===============================  α׷ ==============================
// [μ]: void
// []: 9600Bps,N,8,1 :: Half Duplex
//         P1.1 : TXD
//         P2.2 : RXD
// ---------------------------------------------------------------------------
int main(void)
{
	cpu_setup();						// cpuʱȭ
	com_set(9600);						//  
	_EINT();						// Enable interrupts

	puts_0("serial test program\r\n");			// string tx
	rx_ready();						//  غ
	while(1){
		if((CCTL0 & CCIE) == 0){			//  Ϸ
			putchar_0(RXData);			// ۽
			rx_ready();				//  غ			
		}
	}
}

// ===================================== CPUʱȭ ============================
// [μ] void
// ----------------------------------------------------------------------------
void cpu_setup(void)
{
unsigned dly;
	
	WDTCTL = WDTPW + WDTHOLD;				// Init watchdog timer(stop)
	
// XTX  : Low Freqency
// DIVx : ACLK(4.096Khz) = LXFTCLK(32.768KHz) / 8
	BCSCTL1 = XT2OFF | DIVA1 | DIVA0;
	for(dly = 0; dly < 0x4000U; dly++) ;			// XTAL  ð 

	Set_DCO();						// DCO 

// DIVx : ACLK = LXFTCLK(32.768KHz) / 1
	BCSCTL1 &= ~(DIVA1 | DIVA0);
}

// =================================== DCO  ===============================
// [μ] void
// ----------------------------------------------------------------------------
void Set_DCO(void)
{
unsigned int Compare, Oldcapture = 0;

// CMx  : Capture on rising edge(01)
// CCISx: CCIxB(01)
// CAP  : Capture Mode(1)
	CCTL2 = CCIS0 + CM0 + CAP;				// Define CCR2, CAP, ACLK
// TASSELx : Timer Clock Source ACLK(01)
// TACLR : TimerA Clear
// MCx   : Continuos mode : The timer counts up to 0xffffh(10)
	TACTL = TASSEL1 + TACLR + MC1;				// SMCLK, continous mode
	while(1){
		while((CCTL2 & CCIFG) != CCIFG);		// Wait until capture occured!
		CCTL2 &= ~CCIFG;				// Capture occured, clear flag
		Compare = CCR2;					// Get current captured SMCLK
		Compare = Compare - Oldcapture;			// SMCLK difference
		Oldcapture = CCR2;				// Save current captured SMCLK
		if(DELTA == Compare){
			break;					// if equal, leave "while(1)"
		}
		else if(DELTA < Compare){			// DCO is too fast, slow it down
			DCOCTL--;
			if(DCOCTL == 0xFF){			// Did DCO role under?
				BCSCTL1--;			// Select next lower RSEL
			}
		}
		else{
			DCOCTL++;
			if(DCOCTL == 0x00){			// Did DCO role over?
				BCSCTL1++;                      // Select next higher RSEL
			}
		}
	}  
	CCTL2 = 0;                              		// Stop CCR2 function
	TACTL = 0;						// Stop Timer_A
}

// ================================   =================================
// [μ] int baud
// Item1: Ʈ (1200,2400,4800,9600Bps)
// [] DCO Freq = 3.6864MHz :: 1200,2400,4800,9600Bps
// ----------------------------------------------------------------------------
void com_set(int baud)
{
	TACTL = TASSEL1+TACLR+MC1;				// Clear TAR, Source = SMCLK, Continous mode
	CCTL0 = OUT;						// TXD Idle as Mark 
	P1SEL = TXD;						// P1.1/TA0 for TXD function
	P2SEL = RXD;						// P2.2/TA0 as RXD input
	P1DIR = TXD;						// TXD output on P1
	switch(baud){
		case 1200 : Bitime_5 = 1536;			// 416.666us bit length 
			    Bitime = 3072;			// 833.333us(1200Bps)
			    break;		
		case 2400 : Bitime_5 = 768;			// 208.333us bit length 
			    Bitime = 1536;			// 416.666us(2400Bps)
			    break;		
		case 4800 : Bitime_5 = 384;			// 104.1666us bit length 
			    Bitime = 768;			// 208.3333us(4800Bps)
			    break;		
		case 9600 : Bitime_5 = 192;			// 50us bit length 
			    Bitime = 384;			// 104.1666us(9600Bps)
			    break;		
	}
}

// =================================  غ ================================
// [μ] void
// ----------------------------------------------------------------------------
void rx_ready(void)
{
	RXBitCnt = 8;						// Load Bit counter
	CCTL0 = SCS+CCIS0+OUTMOD0+CM1+CAP+CCIE;			// Sync, Neg Edge, Capture          
}

// =============================== 1Ʈ ۽ ===============================
// [μ] unsigned char datac
// Item1:  (0 - ffh)
// ----------------------------------------------------------------------------
void putchar_0(unsigned char datac)
{
	TXBitCnt = 10;						// Load Bit counter, 8data + ST/SP
	CCR0 = TAR;						// Current state of TA counter
	CCR0 += Bitime;						// Some time till first bit
	TXData = (unsigned int)datac | 0x100;			// Add mark stop bit to TXData 
	TXData = TXData << 1;					// Add space start bit
	CCTL0 = OUTMOD0+CCIE;					// TXD = mark = idle
	while(CCTL0 & CCIE);					// Wait for TX completion
}

// ================================ Ʈ ۽ ===============================
// [μ] unsigned char dbuf
// Item1:   ڿ
// ----------------------------------------------------------------------------
void puts_0(unsigned char dbuf[])
{
unsigned char ch;
int i = 0;

	ch = dbuf[i++];
	while(ch != 0){
		putchar_0(ch);					// 1 send
		ch = dbuf[i++];
	}
}

// ============================ Timer A0 interrupt ============================
// [μ] void
// ----------------------------------------------------------------------------
#pragma vector=TIMERA0_VECTOR
__interrupt void Timer_A (void)
{
	CCR0 += Bitime;						// Add Offset to CCR0
// RX
	if(CCTL0 & CCIS0){					// RX on CCI0B?
		if(CCTL0 & CAP){				// Capture mode = start bit edge 
			CCTL0 &= ~ CAP;				// Switch from capture to compare mode 
			CCR0 += Bitime_5;
		}
		else{
			RXData = RXData >> 1;           
			if(CCTL0 & SCCI) RXData |= 0x80;	// Get bit waiting in receive latch
			RXBitCnt --;				// All bits RXed?
			if (RXBitCnt == 0){
				CCTL0 &= ~ CCIE;		// All bits RXed, disable interrupt
//				_BIC_SR_IRQ(CPUOFF);		// Clear LPM0 bits from 0(SR)	
			}
		}
	}
// TX
	else{
		if(TXBitCnt == 0) CCTL0 &= ~ CCIE;		// All bits TXed, disable interrupt
		else{
			CCTL0 |=  OUTMOD2;			// TX Space
			if(TXData & 0x01) CCTL0 &= ~ OUTMOD2;	// TX Mark
			TXData = TXData >> 1;
			TXBitCnt --;
		}
	}
}
